#ifndef POLYNOMIAL_H
#define POLYNOMIAL_H
/** 
*       This program has been developed by Michal Roy.  
*	You have the author's permission to incorporate this code 
*       into your product, royalty free.  

*       The author specifically disclaims all warranties, express or
*       implied, and all liability, including consequential and
*       other indirect damages, for the use of this code, 
*       including liability for infringement of any proprietary
*       rights, and including the warranties of merchantability
*       and fitness for a particular purpose.  And he does not 
*       assume any responsibility for any errors which may 
*       appear in this code nor any responsibility to update it.
*
*   (c) 2005, Michael Roy.  tika1966@yahoo.com
*/
/** 
*   @file polynomial.h
*
*   Polynomial class definition
*
*   @ingroup Poly
*   @author Michael Roy <a href="mailto:tika1966@yahoo.com">tika1966@yahoo.com</a>
*   @version 1.0
*
*   @see    polyzero.h\n
*           laguerre.h\n
*           muller.h
*/

#include <vector>
#include <algorithm>
#include <complex>
#include <stdarg.h>
#include "floatspecs.h"

/** @defgroup Poly The Polynomial Templates
*   The Polynomial Templates.
*/
/** @defgroup PolyEval Polynomial evaluation functions
*   Polynomial evaluation functions.
*   @ingroup Poly
*/
/** @defgroup PolyOps Polynomial arithmetics functions
*   Polynomial arithmetics functions.
*   @ingroup Poly
*/
/** @addtogroup Poly 
*   @{ */
/**
*   defines operations on polynomials.
*
*   Class Polynomial is based on the class \p std::vector, and thus 
*   requires the Standard C++ Library (STL).
*
*   Coeficients are stored with lowest degree coeficent first, i.e.: the 
*   polynomial \f${a_0}+{a_1}x+{a_2}x^2+{a_3}x^3\f$ will be stored 
*   in the vector as \f$[a_0, a_1, a_2, a_3]\f$.
*
*   The coeficient \f$a_0\f$ of degree 0 is always populated.  That means
*   \p P[0] will always be valid for a polynomial \p P.
*
*   Evaluations are made using Horner's recurence, or synthetic division.
*
*   Valid types for template parameter \e T are any floating point,
*   or complex types which define operators +, -, *, /, and functions abs().
*
*   Several constructors are available, the following example shows them and 
*   their usage.
*
*   @par Example:
*   @include constructor.cpp
*
*   @author Michael Roy <a href="mailto:tika1966@yahoo.com">tika1966@yahoo.com</a>
*   @version 0.1
*
*   @see    polynomial.h for general support function\n
*           polyzero.h for polynomial roots support funtions\n
*           laguerre.h for roots-finding by Laguerre's method\n
*           muller.h for roots-finding by Muller's method
*/

template <class T>
class Polynomial : public std::vector<T> {

public:
    /** @name Construction */
    /** @{ */
    /**
    *   default constructor.
    *
    *   Creates an empty polynomial of degree zero, you can optionally specify
    *   a scalar value.
    *
    *   @param[in]  a   default scalar value for coeficient \f$a_0\f$.
    */

    explicit Polynomial(const T& a = T(0)) {
        push_back(a);
    }

    /**
    *   copy constructor.
    *
    *   Creates an exact copy of Polynomial \f$p\f$
    *
    *   @param[in]  p   source polynomial.
    *
    */

    explicit Polynomial(const std::vector<T>& p) : std::vector(p) { }

    /**
    *   initialization constructor.
    * 
    *   Initializes a polynomial of arbitrary size and coeficients.  Care must be taken
    *   to properly cast the variable arguments to the same type as template type \e T.
    *
    *   @param[in]  n       polynomial's degree
    *   @param[in]  a0      degree 0 coeficient
    *   @param[in]  ...     coeficients \f$a_1..a_n\f$, number of parameters must match 
    *                       parameter \a n.
    */

    Polynomial(unsigned n, T a0, ...) {
        va_list marker;
        int i = n;
        if (sizeof(T) != sizeof(float)) {
            push_back(a0);
            va_start(marker, a0);
            while (i-- > 0) {
                T an = va_arg(marker, T);
                push_back(an);
            }
            va_end(marker);
            compact();
        }
        else {
            // in c/c++, floats are pushed as doubles
            push_back(a0);
            va_start(marker, a0);
            while (i-- > 0) {
                double an = va_arg(marker, double);
                push_back(T(an));
            }
            va_end(marker);
            compact();
        }
    }
    /** @} construction */

    /** @name Methods */
    /** @{ */
    /**
    *   compacts array.
    *
    *   Compacts array size to match actual polynomial degree by enforcing 
    *   the rule <i>p[p.size() - 1] != 0</i>.
    *
    *   This function is called by other member functions to enforce the rule
    *   automatically whenever operations are performed on the polynomial.
    *
    *   @see degree()
    */

    void compact() {
        while((back() == T(0)) && size() > 1) pop_back();
    }
    
    /**
    *   returns degree of polynomial.
    *
    *   Degree of the polynomial is the index of the highest degree non_null
    *   coeficient.
    *
    *   @par Example:
    *   @include degree.cpp
    *   @par Program output:
    *   @include degree.txt
    *
    *   @returns the degree of the polynomial.
    *
    *   @see degree()
    */

    int degree() const {
        int m = (int)size() - 1;
        while ((m > 0) &&( (*this)[m] == T(0))) --m;
        return m;
    }

    /**
    *   returns degree of polynomial.
    *
    *   Degree of the polynomial is the index of the highest degree non_null
    *   coeficient.
    *
    *   This non-constant version of the function calls compact() before 
    *   returning.
    *
    *   @returns The degree of the polynomial.
    *
    *   @see degree() const
    */

    int degree() {
        compact();
        int m = (int)size() - 1;
        if (m < 0) {
            resize(1);
            m = 0;
        }
        return (int)size() - 1;
    }

    /**
    *   shifts the coeficients.
    *
    *   Shifts the polynomial coeficients left or right, effectively 
    *   doing the operation \f$q(x)=x^{n}p(x)\f$.
    *
    *   If \a n is negative, the remainder of the division 
    *   \f$\frac{p(x)}{x^{-n}}\f$ is lost.
    *
    *   @param[in]  n   power of multipler \f$x^n\f$.
    */

    void shift(int n) {
        if (n < 0) {
            int newSize = (int)size() + n;
            if (newSize >= 1) {
                std::copy(&(*this)[-n], &*end(), &(*this)[0]);
            }
            else {
                newSize = 1;
                (*this)[0] = T(0);
            }
            resize(newSize);
        }
        else if (n > 0) {
            resize(size() + n);
            std::copy(begin(), end(), begin() + n);
        }
    }

    /**
    *   makes polynomial monic.
    *
    *   Makes polynomial monic. A polynomial is said monic if its highest 
    *   degree coeficient equals 1.
    *
    *   @see isMonic() const
    */

    void makeMonic() {
        compact();
        if (!isMonic())
            *this /= (*this)[degree()];
    }

    /**
    *   indicates if polynomial is monic.
    *   Indicates if polynomial is monic. A polynomial is said monic if its 
    *   highest degree coeficient equals 1.
    *
    *   @returns \p true if polynomial is monic.
    *
    *   @see makeMonic()
    */

    inline bool isMonic() const {
        return ((*this)[degree()] == T(1));
    }
 
    /**
    *   get derivative of \f$p(x)\f$.
    *
    *  Computes the first derivative polynomial \f$p'(x)\f$ of polynomial \f$p(x)\f$. 
    *
    *   @param[out] pp    on exit, contains the first derivative \f$p'(x)\f$.
    *
    *   @see evalAndDerive(const T& x, T& ppx) const
    */

    void getDerivative(Polynomial<T>& pp) const {
        unsigned i;
        T ii;
        pp.resize(size() - 1);
        for (i = 0, ii = T(1); i < pp.size(); ++i, ii += T(1))
            pp[i] = ii * (*this)[i + 1];
    }

    /** @} methods */

    /** @name Evaluation */
    /** @{ */
    /**
    *   evaluates polynomial value \f$p(x)\f$.
    *
    *   Evaluates \f$p(x)\f$ by Horner's recurence.
    *
    *   No error checking is performed, so any validation for overflow, 
    *   or underflow is the responsibility of the caller.
    *
    *   @param[in]  x   parameter of \f$p(x)\f$
    *
    *   @returns The value of \f$p(x)\f$.
    */

    T eval(const T& x) const {
        return ::eval(*this, x);
    }

    /**
    *   evaluates polynomial value \e p(a), and computes \f$\frac{p(x)}{(x - a)}\f$.
    *
    *   Evaluates \f$p(a)\f$ and polynomial \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$ at the same time 
    *   by Horner's recurence.
    *
    *   It can be shown that \f$p(x)=q(x)(x - a)+p(a)\f$, 
    *   which essentially means that \f$p(a)\f$ is the remainder of 
    *   \f$\frac{p(x)}{(x - a)}\f$. For the special case where \f$a\f$ is a root of 
    *   \f$p(x)\f$ (i.e. \f$p(a)=0\f$), this operation is called 
    *   <i>deflation</i>.  That is, on exit \f$q(x)\f$ will have all the roots of 
    *   \f$p(x)\f$, except for the root \f$a\f$.
    *
    *   No error checking is performed, so any validation for overflow, 
    *   or underflow is the responsibility of the caller.
    *
    *   @param[in]  a   parameter of \f$p(a)\f$.
    *   @param[out] q   on exit contains the quotient \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$.
    *   @returns \f$p(a)\f$, the remainder of \f$\frac{p(x)}{(x - a)}\f$.
    *
    *   @see evalAndDeflate(const Polynomial<T>& p, const U& x, Polynomial<U>& q, V& e)
    */

    T evalAndDeflate(const T& a, Polynomial<T>& q) const {
        return ::evalAndDeflate(*this, a, q);
    }

    /**
    *   evaluates \f$p(x)\f$ and derivate \f$dy=p'(x)\f$ at the same time.
    *
    *   Evaluates efficiently \f$p(x)\f$ and first derivative \f$p'(x)\f$
    *   at the same time.
    *
    *   @param[in]  x   variable to evaluate.
    *   @param[out] ppx on exit contains \f$p'(x)\f$.
    *   @returns The value of \f$p(x)\f$.
    */

    T evalAndDerive(const T& x, T& ppx) const {
        int m = degree();
        T y;
        ppx = 0;
        y = (*this)[m];
        for (int i = m - 1; i >= 0; --i) {
            ppx = (ppx * x) + y;
            y = (y * x) + (*this)[i];
        }
        return y;
    }
    /** @} evaluation */

    /** @name Operators */
    /** @{ */
    /**
    *   make a copy of a polynomial.
    *
    *   Make a copy of a polynomial.
    *
    *   @param[in]  q   polynomial to copy.
    *
    *   @returns A const reference on this object.
    */

    inline const Polynomial<T>& operator= (const std::vector<T>& q) {
        (std::vector<T>&)*this = q;
        return *this;
    }

    /**
    *   in-place polynomial addition.
    *
    *   Performs the addition \f$r=p+q\f$, storing the result stored in \p this.
    *
    *   @param[in]  q   a polynomial.
    *   @returns        A const reference to \p this.
    *
    *   @see    operator+ (const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial::operator-= (const Polynomial<T>& A)\n
    *           add(Polynomial<T>& p, const Polynomial<T>& q)
    */

    inline const Polynomial<T>& operator+= (const Polynomial<T>& q) {
        add(*this, q);
        return *this;
    }

    /**
    *   in-place polynomial substraction.
    *   Performs the substraction \f$r=p-q\f$ and stores the resulting polynomial in 
    *   \p this.
    *
    *   @param[in]  q   a polynomial.
    *   @returns        A const reference to \p this.
    *
    *   @see    operator- (const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial::operator+= (const Polynomial<T>& A)\n
    *           sub(Polynomial<T>& p, const Polynomial<T>& q)
    */

    inline const Polynomial<T>& operator-= (const Polynomial<T>& q) {
        sub(*this, q);
        return *this;
    }

    /**
    *   in-place polynomial multiply.
    *
    *   Multiplies two polynomials by convolution.  Performs the operation 
    *   \f$p\leftarrow pq\f$
    *
    *   @param[in]  p   a polynomial to multiply to \p this object.
    *   @returns        A const reference to this.
    *
    *   @see    mul(Polynomial<T> & r, const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial::operator* (const Polynomial<T>& p, const Polynomial<T>& q)
    */

    inline const Polynomial<T> operator*= (const Polynomial<T>& p) {
        Polynomial<T> q = *this;
        mul(*this, p, q);
        return *this;
    }

    /**
    *   in-place polynomial scaling.
    *
    *   Polynomial scaling, performs the operation \f$p(x)\leftarrow s.p(x)\f$.
    *
    *   @returns A constant reference to \p this.
    *
    *   @see    mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p)\n
    *           Polynomial<T>::operator*= (const T& f)\n
    *           operator/ (const T& f, const Polynomial<T>& p)
    */

    const Polynomial<T>& operator*= (const T& f) {
        mul(*this, f);
        return *this;
    }

    /**
    *   in-place polynomial scaling.
    *
    *   Polynomial scaling, performs the operation \f$p(x)\leftarrow \frac{p(x)}{s}\f$.
    *
    *   @returns A constant reference to \p this.
    *
    *   @see    mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p)\n
    *           Polynomial<T>::operator*= (const T& s)\n
    *           operator/ (const Polynomial<T>& p, const T& s)
    */

    Polynomial<T>& operator/= (const T& s) {
        mul(*this, T(1) / s);
        return *this;
    }
    /** @} operators */

    /** @name Non-member operators */
    /** @{ */
    /**
    *   out-of-place polynomial addition.
    *
    *   Performs the addition \f$r=p+q\f$.
    *
    *   Calling directly the three parameters function add() is actually more 
    *   efficient, as it will avoid an unnecessary copy of the polynomial.
    *
    *   @see    operator- (const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial<T>::operator+= (const Polynomial<T>& p)\n
    *           add(Polynomial<T>& r, const Polynomial<T>& p, const Polynomial<T>& q)
    */

    friend inline Polynomial<T> operator+ (const Polynomial<T>& p, const Polynomial<T>& q) {
        Polynomial<T> r;
        add(r, p, q);
        return r;
    }

    /**
    *   out-of-place polynomial substraction.
    *
    *   Performs the substraction \f$r=p-q\f$.
    *
    *   @param[in]  p   a polynomial.
    *   @param[in]  q   another polynomial.
    *   @returns    The result @e r of the operation \f$r=p-q\f$.
    *
    *   @see    operator+ (const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial::operator-= (const Polynomial<T>& A)\n
    *           add(Polynomial<T>& p, const Polynomial<T>& q)
    */

    friend inline Polynomial<T> operator- (const Polynomial<T>& p, const Polynomial<T>& q) {
        Polynomial<T> r;
        sub(r, a, b);
        return r;
    }

    /**
    *   out-of-place polynomial multiply.
    *
    *   Multiplies two polynomials by convolution.  It is advisable to use 
    *   mul() instead, as it avoids the copying of the resulting polynomial
    *   off the stack.
    *
    *   @param[in]  p   a polynomial.
    *   @param[in]  q   another polynomial.
    *   @returns        The product \f$r=pq\f$
    *
    *   @see    mul(Polynomial<T> & r, const Polynomial<T>& p, const Polynomial<T>& q)\n
    *           Polynomial::operator*= (const Polynomial<T>& p)
    */

    friend inline Polynomial<T> operator* (const Polynomial<T>& p, const Polynomial<T>& q) {
        Polynomial<T> r;
        mul(r, p, q);
        return r;
    }

    /**
    *   out-of_place polynomial scaling.
    *
    *   Polynomial scaling.
    *
    *   @returns The polynomial \f$r(x)=s.p(x)\f$.
    *
    *   @see    mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p)\n
    *           Polynomial<T>::operator*= (const T& s)\n
    *           operator/ (const T& s, const Polynomial<T>& p)
    */

    friend inline Polynomial<T> operator* (const T& s, const Polynomial<T>& p) {
        Polynomial<T> r;
        mul(r, s, p);
        return r;
    }

    /**
    *   out-of-place polynomial scaling.
    *
    *   Polynomial scaling, performs the operation \f$r(x)=\frac{p(x)}{s}\f$.
    *
    *   @returns A constant reference to \p this.
    *
    *   @see    mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p)\n
    *           Polynomial<T>::operator*= (const T& s)\n
    *           operator/= (const T& s)
    */

    friend Polynomial<T> operator/ (const Polynomial<T>& p, const T& s) {
        Polynomial<T> r;
        mul(r, T(1) / f, p);
        return r;
    }
    /** @} Non-member operators */
};

/**
*   out-of-place polynomial addition.
*
*   Performs addition of two polynomials \f$r=p+q\f$, 
*   resulting polynomial is stored in user-supplied polynomial \a r.
*
*   @param[out] r   on exit, contains the polynomial \f$r=p+q\f$.
*   @param[in]  p   first input polynomial.
*   @param[in]  q   second input polynomial.
*
*   @ingroup PolyOps
*   @see    operator+ (const Polynomial<T>& p, const Polynomial<T>& q)\n
*           add(Polynomial<T>& p, const Polynomial<T>& q)
*/

template <class T>
void add(Polynomial<T>& r, const Polynomial<T>& p, const Polynomial<T>& q) {
    int i, m, n;
    m = p.degree();
    n = q.degree();
    if (m <= n) {
        r.resize(n + 1);
        for (i = 0; i <= m; ++i)
            r[i] = p[i] + q[i];
        for ( ; i <= n; ++i)
            r[i] = q[i];
    }
    else {
        r.resize(m + 1);
        for (i = 0; i <= n; ++i)
            r[i] = p[i] + q[i];
        for ( ; i <= m; ++i)
            r[i] = p[i];
    }
    r.compact();
}

/**
*   in-place polynomial addition.
*
*   Performs in-place addition of two polynomials \f$p\leftarrow p+q\f$.
*
*   @param[in,out]  p   first input polynomial.  On exit, contains the sum \f$p+q\f$.
*   @param[in]      q   second input polynomial.
*
*   @ingroup PolyOps
*   @see    add(Polynomial<T>& r, const Polynomial<T>& p, const Polynomial<T>& q)\n
*           Polynomial::operator += (const Polynomial<T>& A)
*/

template <class T>
void add(Polynomial<T>& p, const Polynomial<T>& q) {
    int m = p.degree();
    int n = q.degree();
    if (m < n)
        p.resize(n + 1, T(0));

    for (int i = 0; i <= n; ++i)
        p[i] += q[i];
    p.compact();
}

/**
*   out-of-place substraction of polynomials.
*
*   Performs substraction of two polynomials \f$r=p-q\f$, 
*   resulting polynomial is stored in user-supplied polynomial \a r.
*
*   @param[out] r   on exit, contains the polynomial \f$p-q\f$.
*   @param[in]  p   first input polynomial.
*   @param[in]  q   second input polynomial.
*
*   @ingroup PolyOps
*/

template <class T>
void sub(Polynomial<T>& r, const Polynomial<T>& p, const Polynomial<T>& q) {
    int i;
    int m = p.degree();
    int n = q.degree();
    if (m <= n) {
        r.resize(n + 1);
        for (i = 0; i <= m; ++i)
            r[i] = p[i] - q[i];
        for ( ; i <= n; ++i)
            r[i] = -q[i];
    }
    else {
        r.resize(m + 1);
        for (i = 0; i <= n; ++i)
            r[i] = p[i] - q[i];
        for ( ; i <= m; ++i)
            r[i] = p[i];
    }
    r.compact();
}

/**
*   in-place substraction of polynomials.
*
*   Performs in-place substraction of two polynomials \f$p\leftarrow p-q\f$.
*
*   @param[in,out]  p   first input polynomial.  On exit, contains the difference \f$p-q\f$.
*   @param[in]      q   second input polynomial.
*
*   @ingroup PolyOps
*/

template <class T>
void sub(Polynomial<T>& p, const Polynomial<T>& q) {
    int i;
    int m = p.degree();
    int n = q.degree();
    if (m < n)
        p.resize(n + 1, T(0));
    for (i = 0; i <= n; ++i)
        p[i] -= q[i];
    p.compact();
}

/**
*   out-of-place polynomial division.
*
*   Performs polynomial division.  On exit:
*   \f[q=\frac{u}{v}\f]
*   \f[r=u\%v\f]
*
*   @param[out] q   quotient \f$q=\frac{u}{v}\f$
*   @param[out] r   remainder \f$r=u\%v\f$
*   @param[in]  u   numerator.
*   @param[in]  v   divisor.
*
*   @ingroup PolyOps
*   @see div(Polynomial<T>& u, Polynomial<T>& r, const Polynomial<T>& v)
*/

template <class T>
void div(Polynomial<T>& q, Polynomial<T>& r, const Polynomial<T>& u, const Polynomial<T>& v) {
    int na = u.degree();
    int nb = v.degree();
    if (na >= nb) {
        r = u;
        T t;
        q.resize(1 + na - nb);
        std::fill(&q[0], &q[q.size()], T(0));
        for (int i = na - nb; i >= 0; --i) {
            q[i] = r[nb + i] / v[nb];
            r[nb + i] = T(0);
            for (int j = nb + i - 1; j >= i; --j) {
                r[j + 1] -= (q[i] * v[j - i]);
            }
            r[i + nb] = 0;
        }
        r.compact();
    }
    else {
        r = u;
        q.resize(1);
        q[0] = T(0);
    }
}

/**
*   in-place polynomial division.
*
*   Performs polynomial division.  On exit:
*   \f[u\leftarrow \frac{u}{v}\f]
*   \f[r=u\%v\f]
*
*   @param[in,out]  u   numerator, holds quotient \f$\frac{u}{v}\f$ on exit.
*   @param[out]     r   remainder \f$r=u\%v\f$
*   @param[in]      v   divisor.
*
*   @ingroup PolyOps
*   @see div(Polynomial<T>& q, Polynomial<T>& r, const Polynomial<T>& u, const Polynomial<T>& v)
*/

template <class T>
void div(Polynomial<T>& u, Polynomial<T>& r, const Polynomial<T>& v) {
    int na = u.degree();
    int nb = v.degree();
    if (na >= nb) {
        r = u;
        u.resize(na - nb + 1);
        for (int i = na - nb; i >= 0; --i) {
            u[i] = r[nb + i] / v[nb];
            r[nb + i] = T(0);
            for (int j = nb + i - 1; j >= i; --j)
                r[j] -= (u[i] * v[j - i]);
        }
        r.compact();
    }
    else {
        r = u;
        u.resize(1);
        u[0] = T(0);
    }
}

/**
*   out-of-place polynomial scaling.
*
*   Polynomial scaling. Performs the operation \f$r(x)=s.p(x)\f$.
*
*   @param[out] r   scaled polynomial \f$r(x)=s.p(x)\f$
*   @param[in]  s   scalar.
*   @param[in]  p   A polynomial.
*
*   @ingroup PolyOps
*   @see    mul(Polynomial<T>& p, const T& s)\n
*           operator* (const T& s, const Polynomial<T>& p)\n
*           operator* (const Polynomial<T>& p, const T& s)
*/

template <class T>
void mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p) {
    if (s != T(1)) {
        r.resize(p.degree() + 1);
        for (unsigned i = 0; i < r.size(); ++i)
            r[i] = s * p[i];
    }
}

/**
*   in-place polynomial scaling.
*
*   Polynomial scaling. Performs the operation \f$p(x)\leftarrow s.p(x)\f$.
*
*   @param[in, out] p   A polynomial, on exit contains the product \f$s.p(x)\f$
*   @param[in]      s   scalar.
*
*   @ingroup PolyOps
*   @see    mul(Polynomial<T>& r, const T& s, const Polynomial<T>& p)\n
*           Polynomial<T>::operator*= (Polynomial<T>& p, const T& s)
*/

template <class T>
inline void mul(Polynomial<T>& p, const T& s) {
    mul(p, s, p);
}

/**
*   out-of-place polynomial multiply.
*
*   Multiplies two polynomials by convolution.
*
*   @param[out] r   resulting polynomial \f$r=pq\f$
*   @param[in]  p   a polynomial.
*   @param[in]  q   another polynomial.
*
*   @ingroup PolyOps
*   @see    operator* (const Polynomial<T>& p, const Polynomial<T>& q)\n
*           Polynomial<T>::operator*= (const Polynomial<T>& p)
*/

template <class T>
void mul(Polynomial<T>& r, const Polynomial<T>& p, const Polynomial<T>& q) {
    int na = p.degree();
    int nb = q.degree();
    r.resize(na + nb + 1);
    std::fill(&r[0], &r[r.size()], T(0));
    for (int i = 0; i <= na; ++i)
        for (int j = 0; j <= nb; ++j)
            r[i + j] += p[i] * q[j];
}

/**
*   evaluates polynomial value \f$y = p(x)\f$.
*
*   Evaluates \f$p(x)\f$ by Horner's recurence.  This is the preferred way of 
*   evaluating polynomials, since the member function Polynomial<T>::eval() is
*   resricted to same-type polynomial coeficients and variable \a x
*
*   No error checking is performed, so any validation for overflow, 
*   or underflow is the responsibility of the caller.
*
*   @param[in]  p   a polynomial.
*   @param[in]  x   parameter of \f$p(x)\f$
*   @returns The value of \f$p(x)\f$.
*
*   @ingroup PolyEval
*   @see    Polynomial<T>::eval(const T& x) const\n
*           eval(const Polynomial<T>& p, const U& x, V& e)
*/

template<class T, class U>
U eval(const Polynomial<T>& p, const U& x) {
    if (x == U(0)) 
        return p[0];
    else {
        int i, m = p.degree();
        register U px = p[m];
        for (i = m - 1; i >= 0; --i) {
            px *= x;
            px += p[i];
        }
        return px;
    }
}

/**
*   evaluates polynomial value \f$y = p(x)\f$.
*
*   Evaluates \f$p(x)\f$ by Horner's recurence.  This is the preferred way of 
*   evaluating polynomials, since the member function Polynomial<T>::eval() is
*   resricted to same-type polynomial coeficients and variable \a x
*
*   No error checking is performed, so any validation for overflow, 
*   or underflow is the responsibility of the caller.
*
*   @param[in]  p   a polynomial.
*   @param[in]  x   parameter of \f$p(x)\f$
*   @param[out] e   round-off error to be expected in the calculation of 
*                   \f$p(x)\f$
*   @returns The value of \f$p(x)\f$.
*
*   @ingroup PolyEval
*   @see    eval(const Polynomial<T>& p, const U& x)\
*           Polynomial<T>::eval(const T& x) const
*/

template<class T, class U, class V>
U eval(const Polynomial<T>& p, const U& x, V& e) {
    static FloatSpecs<V> fpSpecs;
    if (x == U(0)) {
        e = V(0);
        return p[0];
    }
    else {
        int i, m = p.degree();
        if (m >= 1) {
            U px = p[m];
            V mx, mpx;
            mx = abs(x);
            e = (abs(px) * fpSpecs.MRE) / (fpSpecs.ARE + fpSpecs.MRE);
            for (i = m - 1; i >= 0; --i) {
                px *= x;
                px += p[i];
                //px = (px * x) + p[i];
                mpx = abs(px);
                e = (mx * e) + mpx;
            }
            e = (e * (fpSpecs.ARE + fpSpecs.MRE)) - (mpx * fpSpecs.MRE);
            return px;
        }
        else {
            e = V(0);
            return p[0];
        }
    }
}

/**
*   evaluates polynomial value \e p(a), and computes \f$\frac{p(x)}{(x - a)}\f$.
*
*   Evaluates \f$p(a)\f$ and polynomial \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$ at the same time 
*   by Horner's recurence.
*
*   It can be shown that \f$p(x)=q(x)(x - a)+p(a)\f$, 
*   which essentially means that \f$p(a)\f$ is the remainder of 
*   \f$\frac{p(x)}{(x - a)}\f$. For the special case where \f$a\f$ is a root of 
*   \f$p(x)\f$ (i.e. \f$p(a)=0\f$), this operation is called 
*   <i>deflation</i>.  That is, on exit \f$q(x)\f$ will have all the roots of 
*   \f$p(x)\f$, except for the root \f$a\f$.
*
*   No error checking is performed, so any validation for overflow, 
*   or underflow is the responsibility of the caller.
*
*   @param[in]  p   a polynomial.
*   @param[in]  a   parameter of \f$p(a)\f$.
*   @param[out] q   on exit contains the quotient \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$.
*   @returns \f$p(a)\f$, the remainder of \f$\frac{p(x)}{(x - a)}\f$.
*
*   @ingroup PolyEval
*   @see    Polynomial<T>::evalAndDeflate(const U& x, Polynomial<U>& q) const\n
*           Polynomial<T>::evalAndDeflate(const U& x, Polynomial<U>& q, V& e) const\n
*           evalAndDeflate(const Polynomial<T>& p, const U& x, Polynomial<U>& q, V& e)
*/

template<class T, class U>
U evalAndDeflate(const Polynomial<T>& p, const U& a, Polynomial<U>& q) {
    int i, m = p.degree();
    if (m > 0) {
        q.resize(m);
        if (a == T(0)) {
            std::copy(&p[1], &p[m + 1], &q[0]);
            return p[0];
        }
        else {
            U pa;
            pa = p[m];
            for (i = m - 1; i >= 0; --i) {
                q[i] = pa;
                pa *= a;
                pa += p[i];
                //pa = (pa * a) + p[i];
            }
            return pa;
        }
    }
    else {
        q.resize(1);
        q[0] = 0;
        return p[0];
    }
}

/**
*   evaluates polynomial value \e p(a), and computes \f$\frac{p(x)}{(x - a)}\f$.
*
*   Evaluates \f$p(a)\f$ and polynomial \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$ at the same time 
*   by Horner's recurence.
*
*   It can be shown that \f$p(x)=q(x)(x - a)+p(a)\f$, 
*   which essentially means that \f$p(a)\f$ is the remainder of 
*   \f$\frac{p(x)}{(x - a)}\f$. For the special case where \f$a\f$ is a root of 
*   \f$p(x)\f$ (i.e. \f$p(a)=0\f$), this operation is called 
*   <i>deflation</i>.  That is, on exit \f$q(x)\f$ will have all the roots of 
*   \f$p(x)\f$, except for the root \f$a\f$.
*
*   No error checking is performed, so any validation for overflow, 
*   or underflow is the responsibility of the caller.
*
*   @param[in]  p   a polynomial.
*   @param[in]  a   parameter of \f$p(a)\f$.
*   @param[out] q   on exit contains the quotient \f$q(x)=\frac{p(x)-p(a)}{(x - a)}\f$.
*   @param[out] e   round-off error to be expected in the calculation of 
*                   \f$p(a)\f$
*   @returns \f$p(a)\f$, the remainder of \f$\frac{p(x)}{(x - a)}\f$.
*
*   @ingroup PolyEval
*   @see    Polynomial<T>::evalAndDeflate(const U& x, Polynomial<U>& q) const\n
*           Polynomial<T>::evalAndDeflate(const U& x, Polynomial<U>& q, V& e) const\n
*           evalAndDeflate(const Polynomial<T>& p, const U& x, Polynomial<U>& q)
*/

template<class T, class U, class V>
U evalAndDeflate(const Polynomial<T>& p, const U& a, Polynomial<U>& q, V& e) {
    static FloatSpecs<V> fpSpecs;
    int i, m = p.degree();
    if (m > 0) {
        q.resize(m);
        if (a == T(0)) {
            e = 0;
            std::copy(&p[1], &p[m + 1], &q[0]);
            return p[0];
        }
        else {
            U pa;
            V mx, mpa;
            pa = p[m];
            mx = abs(a);
            e = (abs(pa) * fpSpecs.MRE) / (fpSpecs.ARE + fpSpecs.MRE);
            for (i = m - 1; i >= 0; --i) {
                q[i] = pa;
                pa *= a;
                pa += p[i];
                mpa = abs(pa);
                e = (mx * e) + mpa;
            }
            e = (e * (fpSpecs.ARE + fpSpecs.MRE)) - (mpa * fpSpecs.MRE);
            return pa;
        }
    }
    else {
        e = 0;
        q.resize(1);
        q[0] = 0;
        return p[0];
    }
}

/**
*   evaluates round-off error.
*
*   Evaluates round_off error in computing \f$p(x)\f$.
*
*   @param[in]  p   a polynomial.
*   @param[in]  mx  absolute value of x.
*   @returns    e   round-off error to be expected in the calculation of 
*                   \f$p(x)\f$
*
*   @ingroup PolyEval
*   @see    eval(const Polynomial<T>& p, const U& x, V& e)\n
*           evalAndDeflate(const Polynomial<T>& p, const U& x, Polynomial<U>& q, V& e)\n
*           Polynomial<T>::evalAndDeflate(const U& x, Polynomial<U>& q, V& e) const
*/

template <class T, class U>
U evalError(const Polynomial<T>& p, const U& mx) {
    static FloatSpecs<U> fpSpecs;
    T px;
    U e, mpx;
    int m = p.degree();
    px = p[m];
    e = (abs(px) * fpSpecs.MRE) / (fpSpecs.ARE + fpSpecs.MRE);
    for (int i = m - 1; i >= 0; --i) {
        px *= mx;
        px += p[i];
        mpx = abs(px);
        e = (mx * e) + mpx;
    }
    return (e * (fpSpecs.ARE + fpSpecs.MRE)) - (mpx * fpSpecs.MRE);
}

/**
*   evaluates \f$p(x)\f$ and first derivate \f$dy=p'(x)\f$ at the same time.
*
*   Evaluates efficiently \f$p(x)\f$ and first derivative \f$p'(x)\f$
*   at the same time.
*
*   @ingroup PolyEval
*   @param[in]  p   a polynomial.
*   @param[in]  x   variable to evaluate.
*   @param[out] ppx on exit contains \f$p'(x)\f$.
*   @returns The value of \f$p(x)\f$.
*
*   @see    evalAndDerive(const Polynomial<T>& p, const U& x, U& ppx, U& pppx)
*/

template <class T, class U>
U evalAndDerive(const Polynomial<T>& p, const U& x, U& ppx) {
    int m = p.degree();
    if (m >= 1) {
        if (x == U(0)) {
            ppx = U(p[1]);
            return U(p[0]);
        }
        else {
            int m = p.degree();
            U px = p[m];
            ppx = U(0);
            for (int i = m - 1; i >= 0; --i) {
                ppx = (ppx * x) + px;
                px = (px * x) + p[i];
                //ppx *= x; 
                //ppx += px;
                //px *= x; 
                //px += p[i];
            }
            return px;
        }
    }
    else {
        ppx = U(0);
        return p[0];
    }
}

/**
*   evaluates \f$p(x)\f$, first derivate \f$p'(x)\f$, and second derivative
*   \f$p''(x)\f$ at the same time.
*
*   Evaluates efficiently \f$p(x)\f$, first derivate \f$p'(x)\f$, and second 
*   derivative \f$p''(x)\f$ at the same time.
*
*   @param[in]  p       a polynomial.
*   @param[in]  x       variable to evaluate.
*   @param[out] ppx     on exit contains \f$p'(x)\f$.
*   @param[out] pppx    on exit contains \f$p''(x)\f$.
*   @returns            The value of \f$p(x)\f$.
*
*   @ingroup PolyEval
*   @see    evalAndDerive(const Polynomial<T>& p, const U& x, U& ppx)
*/

template <class T, class U>
U evalAndDerive(const Polynomial<T>& p, const U& x, U& ppx, U& pppx) {
    int m = p.degree();
    if (m >= 2) {
        if (x == U(0)) {
            ppx = U(p[1]);
            pppx = U(p[2] + p[2]);
            return U(p[0]);
        }
        else {
            int m = p.degree();
            U px = p[m];
            ppx = U(0);
            pppx = U(0);
            for (int i = m - 1; i >= 0; --i) {
                pppx *= x; 
                pppx += ppx;
                ppx *= x; 
                ppx += px;
                px *= x; 
                px += p[i];
            }
            pppx += pppx;
            return px;
        }
    }
    else if (m == 1) {
        ppx = U(p[1]);
        pppx = U(0);
        return (p[1] * x) + p[0];
    }
    else {
        ppx = U(0);
        pppx = U(0);
        return p[0];
    }
}

/**
*   evaluates first derivate \f$p'(x)\f$, second derivative
*   \f$p''(x)\f$, and deflates a polynomial at the same time.
*
*   Evaluates efficiently \f$p(x)\f$, first derivate \f$p'(x)\f$, and second 
*   derivative \f$p''(x)\f$, while deflating the polynomial \a p.
*
*   @param[in]  p       a polynomial.
*   @param[in]  x       variable to evaluate.
*   @param[out] ppx     on exit, contains \f$p'(x)\f$.
*   @param[out] pppx    on exit, contains \f$p''(x)\f$.
*   @param[out] q       on exit, contains \f$q(z)=\frac{p(z)}{z-x}\f$
*   @returns            The value of \f$p(x)\f$.
*
*   @ingroup PolyEval
*   @see    evalAndDerive(const Polynomial<T>& p, const U& x, U& ppx, U& ppx)
*/

template <class T, class U>
U evalDeriveAndDeflate(const Polynomial<T>& p, const U& x, U& ppx, U& pppx, Polynomial<U>& q) {
    int m = p.degree();
    if (m >= 2) {
        if (x == U(0)) {
            q.resize(m);
            std::copy(&p[1], &p[m + 1], &q[0]);
            ppx = U(p[1]);
            pppx = U(p[2] + p[2]);
            return U(p[0]);
        }
        else {
            q.resize(m);
            U px = p[m];
            ppx = U(0);
            pppx = U(0);
            for (int i = m - 1; i >= 0; --i) {
                pppx *= x; 
                pppx += ppx;
                ppx *= x; 
                ppx += px;
                q[i] = px;
                px *= x; 
                px += p[i];
            }
            pppx += pppx;
            return px;
        }
    }
    else if (m == 1) {
        q.resize(1);
        ppx = U(p[1]);
        pppx = U(0);
        q[0] = T(0);
        return (p[1] * x) + p[0];
    }
    else {
        q.resize(1);
        q[0] = U(0);
        ppx = U(0);
        pppx = U(0);
        return p[0];
    }
}

/**
*   returns optimal scale for polynomial.
*
*   Computes an optimal scale factor for polynomial \a p.  Thus ensuring that
*   evaluating the polynomial will not result in overflow or underflow.
*
*   @param[in]  p       a polynomial.
*
*   @see U scalePoly(Polynomial<T>& p, const Polynomial<U>& q);
*/

template<class T>
T getOptimalScale(const Polynomial<T>& p) {
    static FloatSpecs<T> fpSpecs;
    T hi = sqrt(fpSpecs.INFINY);
    T lo = fpSpecs.SMALNO / fpSpecs.ETA;
    T mx = 0;
    T mn = fpSpecs.INFINY;
    T x;
    for (unsigned i = 1; i < p.size(); ++i) {
        if (p[i] > mx) mx = p[i];
        if (p[i] != 0 && p[i] < mn) mn = p[i];
    }

    if (mn >= lo && mx <= hi)
        return T(1);

    x = lo / mn;
    if (x <= T(1))
        x = 1 / (sqrt(mx) * sqrt(mn));
    else
        if (fpSpecs.INFINY / x > mx)
            x = T(1);

    return pow(fpSpecs.BASE, floor(T(.5) + (log(x) / fpSpecs.LOG_BASE)));
}

/**
*   rescales a polynomial using an optimal scale.
*
*   Computes an optimal scale factor for polynomial \a p.  Thus ensuring that
*   evaluating the polynomial will not result in overflow or underflow, then
*   scales the polynomial \a p.
*
*   @param[in,out]  p       a polynomial.
*   @param[in]      q       the modulus polynomial of p.
*
*   @see    scalePoly(Polynomial<T>& p, const Polynomial<U>& q)\n
*           modulus(const Polynomial<T>& p, Polynomial<U>& q)
*/

template <class T, class U>
U scalePoly(Polynomial<T>& p, const Polynomial<U>& q) {
    U scale = getOptimalScale(q);
    p *= scale;
    return scale;
}

/**
*   computes the modulus of polynomial \a p.
*
*   Computes the modulus of polynomial \a p, such as for polynomial:
*
*   \f[p(x)=a_0+a_1{x}+a_2{x^2}+...+a_n{x^n}\f]
*
*   the resulting polynomial \a q will be :
*
*   \f[q(x)=|a_0|+|a_1|{x}+|a_2|{x^2}+...+|a_n|{x^n}\f]
*
*   @param[in]  p       a polynomial.
*   @param[out] q       the modulus polynomial p.
*/

template <class T, class U>
void modulus(const Polynomial<T>& p, Polynomial<U>& q) {
    int m = p.degree();
    q.resize(m + 1);
    for(int i = 0; i <= m; ++i) 
        q[i] = abs(p[i]);
}

/**@} Poly */
#endif //POLYNOMIAL_H
